/*
errmsg.c
last touched in Par 1.53.0
last meaningful change in Par 1.40
Copyright 1993 Adam M. Costello

This is ANSI C code (C89).

*/


#include "errmsg.h"  /* Makes sure we're consistent with the declarations. */


const char * const outofmem =
  "Out of memory.\n";

const char * const impossibility =
  "Impossibility #%d has occurred.  Please report it.\n";
